/* fton.c
 * Copyright (C) 1998 N.Fukase
 * Modified by William Stein (was@math.berkeley.edu), November 1999.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <time.h>
#include <utime.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "rupcommand.h"
#include "rupserial.h"

static char	rup_line[FILENAME_MAX] = {"/dev/ttyS0"};
static char	rup_cwd[64] = {"B:\\"};
static char	rup_cwdA[64] = {"A:\\"};
static char	rup_cwdB[64] = {"B:\\"};
static char	prompt[80]= {"fton> "};
static char	ftr_version[]= {"0.1a"};

/*===== readlineط ========================================================*/
/*_____ ̵Τǥץȥ ________________________________________*/
int	rlc_ls(char **coms);
int	rlc_ver(char **coms);
int	rlc_ren(char **coms);
int	rlc_cp(char **coms);
int	rlc_mkdir(char **coms);
int	rlc_rmdir(char **coms);
int	rlc_del(char **coms);
int	rlc_lpwd(char **coms);
int	rlc_lcd(char **coms);
int	rlc_pwd(char **coms);
int	rlc_put(char **coms);
int	rlc_get(char **coms);
int	rlc_lls(char **coms);
int	rlc_cd(char **coms);
int	rlc_help(char **coms);
/*_____ ޥɥ ______________________________________________________*/
typedef enum {
     ca_none = 0,
     ca_remote,
     ca_remote_d,
     ca_local,
     ca_local_d,
     ca_line,
     ca_command,
     ca_valiable
} COMATTR;
/*_____ ޥɤΥꥹȤι¤ ____________________________________________*/
typedef struct {
     char	*name;
     COMATTR	attr;
     Function	*func;
     char	*doc;
} COMMAND;
/*_____ ޥɤΥꥹ ____________________________________________________*/
COMMAND commands[] = {
     { "cd", ca_remote_d, rlc_cd, "change onHand\'s current working directory"},
     { "close", ca_none, NULL, "close the connection to onHand"},
     { "cp", ca_remote_d, rlc_cp, "copy file in  onHand\'s"},
     { "delete", ca_remote, rlc_del, "delete file in onHand"},
     { "get", ca_remote, rlc_get, "get a file from onHand" },
     { "help", ca_command, rlc_help, "display this text" },
     { "?", ca_command, rlc_help, "display this text"},
     { "lcd", ca_local_d, rlc_lcd, "change local current working directory"},
     { "lls", ca_local, rlc_lls, "list contents of local directory"},
     { "lpwd", ca_none, rlc_lpwd, "print local current working directory" },
     { "ls", ca_remote, rlc_ls, "list contents of onHand\'s directory" },
     { "mkdir", ca_remote, rlc_mkdir, "make directory in onHand" },
     { "mv", ca_remote, rlc_ren, "rename file in onHand" },
     { "open", ca_none, NULL, "open a connection to onHand"},
     { "put", ca_local, rlc_put, "put a file into onHand" },
     { "pwd", ca_none, rlc_pwd, "print the current working directory of onHand"},
     { "quit", ca_none, NULL, "quit this programm"},
     { "rename", ca_remote, rlc_ren, "rename file"},
     { "rm", ca_remote, rlc_del, "remove file in onHand" },
     { "rmdir", ca_remote, rlc_rmdir, "remove a directory in onHand" },
     { "ver", ca_none, rlc_ver, "print WPs-DOS\'s version"},
     { (char *)NULL, ca_none, (Function *)NULL, (char *)NULL }
};
/*_____ readline Ruputerե̾䴰ѥꥹȥͥ졼 __________________*/
char *
ruputerfile_generator(char *text, int state)
{
     static int		list_index, len;
     static RDENT	rde = {0, NULL, ""};

     if (!state) {
	  list_index = -1;
	  len = strlen(text);
	  free(rde.rfi);
	  strcpy(rde.path, rup_cwd);
	  strcat(rde.path, "*.*");
	  command_dir(&rde);
     }
     while ((list_index + 1) < rde.nfiles) {
	  list_index++;
	  if (strncmp(rde.rfi[list_index].name, text, len) == 0) {
	       return strdup(rde.rfi[list_index].name);
	  }
     }
     return ((char *)NULL);
}
/*_____ readlineޥ䴰ѥꥹȥͥ졼 ____________________________*/
char *
command_generator (char *text, int state)
{
     static int	list_index, len;
     char	*name;
     
     if (!state) {
	  list_index = 0;
	  len = strlen(text);
     }
     while ((name = commands[list_index].name) != NULL) {
	  list_index++;
	  if (strncmp (name, text, len) == 0)
	       return strdup(name);
     }
     return ((char *)NULL);
}
/*_____ readline䴰 ________________________________________________________*/
char **
rft_comp(char *text, int start, int end)
{
     char	**matches;
     char	buf[64];
     int	i;
     
     matches = (char **)NULL;
     if (start == 0)
	  matches = completion_matches(text, command_generator);
     else {
	  strcpy(buf, rl_line_buffer);
	  if(strtok(buf, " \t") != NULL) {
	       for(i = 0; commands[i].name != NULL; i++)  {
		    if (strcmp(commands[i].name, buf) == 0) {
			 if((commands[i].attr == ca_remote) || (commands[i].attr == ca_remote_d)) {
			      matches = completion_matches(text, ruputerfile_generator);
			      break;
			 } else if(commands[i].attr == ca_command) {
			      matches = completion_matches(text, command_generator);
			      break;
			 }
		    }
	       }
	  }
     }
     return (matches);
}
/*_____ readline ______________________________________________________*/
void
init_readline(void)
{
     rl_readline_name = "rft";
     rl_attempted_completion_function = (CPPFunction *)rft_comp; 
}
/*_____ 󥿥饯ƥ֥⡼ ______________________________________________*/
void
intteruct(void)
{
     char	*coms[256], *name, *line;
     int	i;

     init_readline();
     for(;;) {
	  line = readline(prompt);
	  if (line && *line) {
	       add_history(line);
	       /* ޥɤȰ򥹥ڡ֤ʬΥ */
	       if((coms[0] = strtok(line, " \t")) == NULL) {
		 free(line);
		 continue;
	       }
	       for (i = 1; (coms[i] = strtok(NULL, " \t")) != NULL; ++i)
		    ;
	       /* ޥɤ򸡺Ƽ¹ */
	       for (i = 0; (name = commands[i].name) != NULL; i++) {
		    if (*coms[0] == 'q') {
			 free(line);
			 return;
		    }
		    if (strcmp(name, coms[0]) == 0) {
			 if(commands[i].func != NULL)
			      (*(commands[i].func))(coms);
			 else
			       printf("Sorry! this command is not available\n");
			 break;
		    }
	       }
	       if (name == NULL)
		    printf("no such command `%s\'\n", coms[0]);
	  }
	  free(line);
     }
}
/*===== rupcomط ==========================================================*/
/*_____ ե ________________________________________________________*/
void
send_file(char *fn, char *rp)
{
     char		*ptr;
     int		fd;
     LENDAT		dat; /* žǡǼ	*/
     RFINFO		rfi; /* žǡ		*/
     struct stat	st;
     struct tm		*ftm;

     if ((fd = open(fn, O_RDONLY)) >= 0) {
	  fstat(fd, &st);
	  if ((dat.data = (char *)malloc(st.st_size)) == NULL) {
	       fprintf(stderr, "Cannot allocate memory\n");
	       close(fd);
	  } else {
	       read(fd, dat.data, st.st_size);
	       dat.length = st.st_size;
	       ftm = localtime(&(st.st_mtime));
	       rfi.cdate = ((((ftm->tm_year - 80) & 0x7f) << 9) | \
			    (((ftm->tm_mon & 0xf) + 1) << 5) | \
			    (ftm->tm_mday & 0x1f));
	       rfi.ctime = (((ftm->tm_hour & 0x1f) << 11) | \
			    ((ftm->tm_min & 0x3f) << 5) | \
			    ((ftm->tm_sec >> 1) & 0x1f));
	       rfi.attr = 0;
	       rfi.size = st.st_size;
	       close(fd);
	       if((ptr = strrchr(fn, '/')) == NULL) {
		    strcpy(rfi.name, fn);
	       }else {
		    strcpy(rfi.name, ptr + 1);
	       }
	       command_upload(&dat, &rfi, rp);
	       free(dat.data);
	  }
     } else
	  fprintf(stderr, "Cannot open file %s\n", fn);
}
/*_____ ե ________________________________________________________*/
void
recv_file(char *fn, char *rp)
{
     int		fd;
     LENDAT		dat; /* žǡǼ	*/
     RFINFO		rfi; /* žǡ		*/
     struct tm		ftm;
     struct utimbuf	ut;

     if ((fd = open(fn, O_RDWR | O_CREAT | O_EXCL, 0777)) >= 0) {
	  strncpy(rfi.name, fn, 12);
	  rfi.name[13] = '\0';
	  if (command_download(&dat, &rfi, rp) != 0) {
	       close(fd);
	       return;
	  }
	  write(fd, dat.data, dat.length);
	  close(fd);
	  ftm.tm_year = ((rfi.cdate >> 9) & 0x7f) + 80;
	  ftm.tm_mon = ((rfi.cdate >> 5) & 0xf) - 1;
	  ftm.tm_mday = (rfi.cdate & 0x1f);
	  ftm.tm_hour = ((rfi.ctime >> 11) & 0x1f);
	  ftm.tm_min = ((rfi.ctime >> 5) & 0x3f);
	  ftm.tm_sec = ((rfi.ctime & 0x1f) << 1);
	  ftm.tm_isdst = 0;
	  ut.actime = mktime(&ftm);
	  ut.modtime = ut.actime;
	  utime(rfi.name, &ut);
	  free(dat.data);
     } else
	  fprintf(stderr, "Already have file %s\n", fn);
}
/*_____ ls _________________________________________________________________*/
void
get_file_list(char *rf, char *rp)
{
     RDENT	rde;
     int	i;

     strcpy(rde.path, rp);
     if (rf == NULL)
	  strcat(rde.path, "*.*");  
     else
	  strcat(rde.path, rf);  
     command_dir(&rde);

     for(i = 0; i < rde.nfiles; ++i) {
	  printf("%s\n", rde.rfi[i].name);
     }
     free(rde.rfi);
}
/*===== ᥤ롼 ======================================================*/
int
main(int argc, char *argv[])
{
     char	*name;
     int	i, c = 1, opts = 1;
     int        rn;  /* ruputer negotiate */

     for (i = 1; i < argc; i++) {
	  name = argv[i];
	  if (*name == '-') {
	       opts++;
	       for (name++; *name != '\0'; name++) {
		    argv[i] = NULL;
		    if (c == i )
			 c++;
		    switch(*name) {
		    case 'V':
			 fprintf(stderr, "%s version %s\n", argv[0], ftr_version);
			 fprintf(stderr, "Copyright (C) 1998 N.Fukase\n");
			 fprintf(stderr, "All Rigth Reserverd\n");
			 return 0;
		    case 'v':
			 fprintf(stderr, "%s version %s\n", argv[0], ftr_version);
			 return 0;
		    case 'l':
			 if (name[1] == '\0') {
			      opts++;
			      i++;
			      strcpy(rup_line, argv[i]);
			      argv[i] = NULL;
			      if (c == i)
				   c++;
			      goto exit_loop;
			 } else {
			      strcpy(rup_line, name);
			      goto exit_loop;
			 }
			 break;
		    default:
			 fprintf(stderr, "Invalid option letter %c\n", *name);
			 break;
		    }
		      
	       }
	  exit_loop:
	  }
     }
     if (argc == opts) {
	  /* ³ */
	  fprintf(stderr, "Connecting onHand on %s....\n", rup_line);
          fprintf(stderr, "Trying to initialize line...\n");
	  init_line(rup_line);
          fprintf(stderr, "Line initialized.\n");
	  while((rn = ruputer_negotiate()) < 0)
	    fprintf(stderr, "Failure: ruputer_negotiate = %d\n", rn);
	  fprintf(stderr, "Connect!\n");
	  command_chdir("b:\\");
	  intteruct();
	  /* λ */
	  command_end();
	  restore_line();
	  return 0;
     } else {
	  for (i = 0; (name = commands[i].name) != NULL; i++) {
	       if ((strcmp(name, argv[c]) == 0) && (commands[i].func != NULL)){
		    init_line(rup_line);
		    while(ruputer_negotiate() < 0)
			 ;
		    command_chdir("b:\\");
		    (*(commands[i].func))(&argv[c]);
		    command_end();
		    restore_line();
		    return 0;
	       }
	  }
	  return -1;
     }
}
/*===== readlineޥ ====================================================*/
/*_____ ls RuputerΥեꥹȤɽ ____________________________________*/
int	rlc_ls(char **coms)
{
     if(coms[1] == NULL)
	  get_file_list("*.*", rup_cwd);
     else
	  get_file_list(coms[1], rup_cwd);
     return 0;
}
/*_____ ver RuputerOSСɽ _____________________________________*/
int	rlc_ver(char **coms)
{
     int	r;

     r = command_ver();
     printf("%d.%d\n", (r & 0xff), ((r & 0xff00) >> 8));
     return 0;
}
/*_____ pwd RuputerCWDɽ ______________________________________________*/
int	rlc_pwd(char **coms)
{
     printf("onHand\'s Current Directory is %s\n", rup_cwd);
     return 0;
}
/*_____ ren RuputerΥե͡ ___________________________________*/
int	rlc_ren(char **coms)
{
     if ((coms[1] != NULL ) & (coms[2] != NULL)) {
	  command_ren(coms[2], coms[1]);
	  return 0;
     } else {
	  printf("Too few arguement\n");
	  return -1;
     }
}
/*_____ pwd RuputerΥե򥳥ԡ _____________________________________*/
int	rlc_cp(char **coms)
{
     if ((coms[1] != NULL ) & (coms[2] != NULL)) {
	  command_copy(coms[2], coms[1]);
	  return 0;
     } else {
	  printf("Too few arguement\n");
	  return -1;
     }
}
/*_____ mkdir Ruputer˥ǥ쥯ȥ ___________________________________*/
int	rlc_mkdir(char **coms)
{
     int	i;

     if (coms[1] != NULL) {
	  for (i = 1; coms[i] != NULL ; ++i) {
	       command_mkdir(coms[i]);
	  }
	  return 0;
     } else {
	  printf("Too few arguement\n");
	  return -1;
     }
}
/*_____ rmdir RuputerΥǥ쥯ȥ ___________________________________*/
int	rlc_rmdir(char **coms)
{
     int	i;

     if (coms[1] != NULL) {
	  for (i = 1; coms[i] != NULL ; ++i) {
	       command_rmdir(coms[i]);
	  }
	  return 0;
     } else {
	  printf("Too few arguement\n");
	  return -1;
     }
}
/*_____ del RuputerΥե _________________________________________*/
int	rlc_del(char **coms)
{
     int	i;

     if (coms[1] != NULL) {
	  for (i = 1; coms[i] != NULL ; ++i) {
	       command_del(coms[i]);
	  }
	  return 0;
     } else {
	  printf("Too few arguement\n");
	  return -1;
     }
}
/*_____ lpwd CWDɽ(lsؤwrapper) _________________________________*/
int	rlc_lpwd(char **coms)
{
     char buf[FILENAME_MAX];

     getcwd(buf, FILENAME_MAX);
     printf("Local current directory is %s\n", buf);
     return 0;
}
/*_____ lcd CWDư _________________________________________________*/
int	rlc_lcd(char **coms)
{
     char buf[FILENAME_MAX];

     if (coms[1] != NULL)
	  chdir(coms[1]);
     else
	  chdir(getenv("HOME"));
     getcwd(buf, FILENAME_MAX);
     printf("Now, local current directory is %s\n", buf);
     return 0;
}
/*_____ put Ruputerإեž ___________________________________________*/
int	rlc_put(char **coms)
{
     if (coms[1] != NULL)
	  send_file(coms[1], rup_cwd);
     return 0;
}
/*_____ get Ruputerեž _______________________________________*/
int	rlc_get(char **coms)
{
     if (coms[1] != NULL)
	  recv_file(coms[1], rup_cwd);
     return 0;
}
/*_____ lls ΥեꥹȤɽ(lsؤwrapper) ___________________*/
int	rlc_lls(char **coms)
{
     int	pid, status;

     if ((pid = fork()) < 0) {
	  perror("fork");
	  return -1;
     }
     if (pid == 0) {
	  strcpy(coms[0], "ls");
	  execvp(coms[0],  coms);
	  perror(coms[0]);
	  exit(-1);
     }
     while(wait(&status) != pid)
	  ;
     return status;
}
/*_____ cd RuputerCWDѹ _______________________________________________*/
int	rlc_cd(char **coms)
{
     char	*frm, *to, *rc, *sl, *nsl, *drv;
     RDENT	rde;

     rc  = NULL;  sl  = NULL;  // to stop the warnings.
     frm = coms[1];
     to = rde.path;

     if (frm == NULL)
	  return rlc_pwd(coms);
     if (frm[1] == ':') {
	  if ((frm[0] == 'A') || (frm[0] == 'a')) {
	       rc = drv = rup_cwdA;
	  } else if((frm[0] == 'B') || (frm[0] == 'b')) {
	       rc = drv = rup_cwdB;
	  } else {
	       printf("Invalid drive letter\n");
	       return -1;
	  }
	  frm += 2;
     } else {
	  if (*rup_cwd == 'A') {
	       rc = drv = rup_cwdA;
	  } else if(*rup_cwd == 'B') {
	       rc = drv = rup_cwdB;
	  }
	  strncpy(to, rc, 2);
     }
     strncpy(to, rc, 2);
     to += 2;
     rc += 2;
     nsl = NULL;
     if ((*frm != '\\') && (*frm != '/')) {
	  /* Хѥ */
	  while((*to = *rc) != '\0') {
	       if ((*to == '\\') || (*to == '/')) {
		    *to = '\\';
		    if(nsl  == NULL)
			 nsl = sl = to;
		    else {
			 sl = nsl;
			 nsl = to;
		    }
	       }
	       to++;
	       rc++;
	  }
	  if (to[-1] != '\\') {
	       *to = '\\';
	       if(nsl  == NULL)
		    nsl = sl = to;
	       else {
		    sl = nsl;
		    nsl = to;
	       }
	       to++;
	  }
     } else {
	  *to = '\\';
	  nsl = sl = to;
	  frm++;
	  to++;
     }
     while(*frm != '\0') {
	  if ((*frm == '.') && (frm[1] == '.')){
	       frm++;
	       if (to[-1] != '\\') 
		    to = nsl;
	       else {
		    to = sl;
		    *to = '\0';
		    nsl = sl;
		    sl = strrchr(rde.path, '\\');
		    *to = '\\';
		    to++;
		    if (sl == NULL)
			 sl = nsl;
	       }
	  } else {
	       if ((*frm == '/') || (*frm == '\\')) {
		    if (to[-1] == '\\') {
			 frm++;
			 continue;
		    }
		    *to = '\\';
		    if(nsl  == NULL)
			 nsl = sl = to;
		    else {
			 sl = nsl;
			 nsl = to;
		    }
		    sl = to;
	       }else
		    *to = *frm;
	       to++;
	  }
	  frm++;
     }
     *to = '\0';
     if((strcmp(rde.path, "A:\\") == 0) || (strcmp(rde.path, "B:\\") == 0)) {
	  printf("Now, onHand\'s current directory is %s\n", rde.path);
	  strcpy(rup_cwd, rde.path);
	  strcat(rup_cwd, "\\");
	  if (*rup_cwd == 'A')
	       strcpy(rup_cwdA, rde.path);
	  else if (*rup_cwd == 'B')
	       strcpy(rup_cwdB, rde.path);
     } else {
	  command_dir(&rde);
	  if(rde.nfiles == 1) {
	       printf("Now, onHand\'s current directory is %s\n", rde.path);
	       strcpy(rup_cwd, rde.path);
	       strcat(rup_cwd, "\\");
	       if (*rup_cwd == 'A')
		    strcpy(rup_cwdA, rde.path);
	       else if (*rup_cwd == 'B')
		    strcpy(rup_cwdB, rde.path);
	  }
     }
     return 0;
}
/*_____ 饤إפɽ ______________________________________________*/
int
rlc_help(char **coms)
{
     int	i;

     if(coms[1] != NULL) {
	  for(i = 0; commands[i].name != NULL; i++)  {
	       if (strcmp(commands[i].name, coms[1]) == 0) {
		    printf("%s\t%s\n", commands[i].name, commands[i].doc);
		    return 0;
	       }
	  }
     }
     for (i = 0; commands[i].name != NULL; ++i) {
	  printf("%s\t%s\n", commands[i].name, commands[i].doc);
     }
     return 0;
}
